@@ -13,6 +13,10 @@
 # define MINGW_HAS_SECURE_API 1
 #endif
 
+#ifndef __STDC_WANT_LIB_EXT1__
+#define __STDC_WANT_LIB_EXT1__ 1 /* for qsort_s() */
+#endif
+
 #include "ruby/internal/config.h"
 
 #include <ctype.h>
@@ -27,6 +31,7 @@
 
 #include "internal.h"
 #include "internal/sanitizers.h"
+#include "internal/imemo.h"
 #include "internal/util.h"
 #include "ruby/util.h"
 #include "ruby_atomic.h"
@@ -45,8 +50,8 @@ ruby_scan_oct(const char *start, size_t len, size_t *retlen)
         if ((s[0] < '0') || ('7' < s[0])) {
             break;
         }
-	retval <<= 3;
-	retval |= *s++ - '0';
+        retval <<= 3;
+        retval |= *s++ - '0';
     }
     *retlen = (size_t)(s - start);
     return retval;
@@ -65,9 +70,9 @@ ruby_scan_hex(const char *start, size_t len, size_t *retlen)
         if (d < 0 || 15 < d) {
             break;
         }
-	retval <<= 4;
-	retval |= d;
-	s++;
+        retval <<= 4;
+        retval |= d;
+        s++;
     }
     *retlen = (size_t)(s - start);
     return retval;
@@ -107,15 +112,15 @@ ruby_scan_digits(const char *str, ssize_t len, int base, size_t *retlen, int *ov
     *overflow = 0;
 
     if (!len) {
-	*retlen = 0;
-	return 0;
+        *retlen = 0;
+        return 0;
     }
 
     do {
-	int d = ruby_digit36_to_number_table[(unsigned char)*str++];
+        int d = ruby_digit36_to_number_table[(unsigned char)*str++];
         if (d == -1 || base <= d) {
-	    --str;
-	    break;
+            --str;
+            break;
         }
         if (mul_overflow < ret)
             *overflow = 1;
@@ -410,57 +415,57 @@ ruby_qsort(void* base, const size_t nel, const size_t size, cmpfunc_t *cmp, void
       register char *m1;
       register char *m3;
       if (n >= 200) {
-	n = size*(n>>3); /* number of bytes in splitting 8 */
-	{
-	  register char *p1 = l  + n;
-	  register char *p2 = p1 + n;
-	  register char *p3 = p2 + n;
-	  m1 = med3(p1, p2, p3);
-	  p1 = m  + n;
-	  p2 = p1 + n;
-	  p3 = p2 + n;
-	  m3 = med3(p1, p2, p3);
-	}
+        n = size*(n>>3); /* number of bytes in splitting 8 */
+        {
+          register char *p1 = l  + n;
+          register char *p2 = p1 + n;
+          register char *p3 = p2 + n;
+          m1 = med3(p1, p2, p3);
+          p1 = m  + n;
+          p2 = p1 + n;
+          p3 = p2 + n;
+          m3 = med3(p1, p2, p3);
+        }
       }
       else {
-	n = size*(n>>2); /* number of bytes in splitting 4 */
-	m1 = l + n;
-	m3 = m + n;
+        n = size*(n>>2); /* number of bytes in splitting 4 */
+        m1 = l + n;
+        m3 = m + n;
       }
       m = med3(m1, m, m3);
     }
 
     if ((t = (*cmp)(l,m,d)) < 0) {                           /*3-5-?*/
       if ((t = (*cmp)(m,r,d)) < 0) {                         /*3-5-7*/
-	if (chklim && nel >= chklim) {   /* check if already ascending order */
-	  char *p;
-	  chklim = 0;
-	  for (p=l; p<r; p+=size) if ((*cmp)(p,p+size,d) > 0) goto fail;
-	  goto nxt;
-	}
-	fail: goto loopA;                                    /*3-5-7*/
+        if (chklim && nel >= chklim) {   /* check if already ascending order */
+          char *p;
+          chklim = 0;
+          for (p=l; p<r; p+=size) if ((*cmp)(p,p+size,d) > 0) goto fail;
+          goto nxt;
+        }
+        fail: goto loopA;                                    /*3-5-7*/
       }
       if (t > 0) {
-	if ((*cmp)(l,r,d) <= 0) {mmswap(m,r); goto loopA;}     /*3-5-4*/
-	mmrot3(r,m,l); goto loopA;                           /*3-5-2*/
+        if ((*cmp)(l,r,d) <= 0) {mmswap(m,r); goto loopA;}     /*3-5-4*/
+        mmrot3(r,m,l); goto loopA;                           /*3-5-2*/
       }
       goto loopB;                                            /*3-5-5*/
     }
 
     if (t > 0) {                                             /*7-5-?*/
       if ((t = (*cmp)(m,r,d)) > 0) {                         /*7-5-3*/
-	if (chklim && nel >= chklim) {   /* check if already ascending order */
-	  char *p;
-	  chklim = 0;
-	  for (p=l; p<r; p+=size) if ((*cmp)(p,p+size,d) < 0) goto fail2;
-	  while (l<r) {mmswap(l,r); l+=size; r-=size;}  /* reverse region */
-	  goto nxt;
-	}
-	fail2: mmswap(l,r); goto loopA;                      /*7-5-3*/
+        if (chklim && nel >= chklim) {   /* check if already ascending order */
+          char *p;
+          chklim = 0;
+          for (p=l; p<r; p+=size) if ((*cmp)(p,p+size,d) < 0) goto fail2;
+          while (l<r) {mmswap(l,r); l+=size; r-=size;}  /* reverse region */
+          goto nxt;
+        }
+        fail2: mmswap(l,r); goto loopA;                      /*7-5-3*/
       }
       if (t < 0) {
-	if ((*cmp)(l,r,d) <= 0) {mmswap(l,m); goto loopB;}   /*7-5-8*/
-	mmrot3(l,m,r); goto loopA;                           /*7-5-6*/
+        if ((*cmp)(l,r,d) <= 0) {mmswap(l,m); goto loopB;}   /*7-5-8*/
+        mmrot3(l,m,r); goto loopA;                           /*7-5-6*/
       }
       mmswap(l,r); goto loopA;                               /*7-5-5*/
     }
@@ -479,18 +484,18 @@ ruby_qsort(void* base, const size_t nel, const size_t size, cmpfunc_t *cmp, void
     loopA: eq_l = 1; eq_r = 1;  /* splitting type A */ /* left <= median < right */
     for (;;) {
       for (;;) {
-	if ((l += size) == r)
-	  {l -= size; if (l != m) mmswap(m,l); l -= size; goto fin;}
-	if (l == m) continue;
-	if ((t = (*cmp)(l,m,d)) > 0) {eq_r = 0; break;}
-	if (t < 0) eq_l = 0;
+        if ((l += size) == r)
+          {l -= size; if (l != m) mmswap(m,l); l -= size; goto fin;}
+        if (l == m) continue;
+        if ((t = (*cmp)(l,m,d)) > 0) {eq_r = 0; break;}
+        if (t < 0) eq_l = 0;
       }
       for (;;) {
-	if (l == (r -= size))
-	  {l -= size; if (l != m) mmswap(m,l); l -= size; goto fin;}
-	if (r == m) {m = l; break;}
-	if ((t = (*cmp)(r,m,d)) < 0) {eq_l = 0; break;}
-	if (t == 0) break;
+        if (l == (r -= size))
+          {l -= size; if (l != m) mmswap(m,l); l -= size; goto fin;}
+        if (r == m) {m = l; break;}
+        if ((t = (*cmp)(r,m,d)) < 0) {eq_l = 0; break;}
+        if (t == 0) break;
       }
       mmswap(l,r);    /* swap left and right */
     }
@@ -498,18 +503,18 @@ ruby_qsort(void* base, const size_t nel, const size_t size, cmpfunc_t *cmp, void
     loopB: eq_l = 1; eq_r = 1;  /* splitting type B */ /* left < median <= right */
     for (;;) {
       for (;;) {
-	if (l == (r -= size))
-	  {r += size; if (r != m) mmswap(r,m); r += size; goto fin;}
-	if (r == m) continue;
-	if ((t = (*cmp)(r,m,d)) < 0) {eq_l = 0; break;}
-	if (t > 0) eq_r = 0;
+        if (l == (r -= size))
+          {r += size; if (r != m) mmswap(r,m); r += size; goto fin;}
+        if (r == m) continue;
+        if ((t = (*cmp)(r,m,d)) < 0) {eq_l = 0; break;}
+        if (t > 0) eq_r = 0;
       }
       for (;;) {
-	if ((l += size) == r)
-	  {r += size; if (r != m) mmswap(r,m); r += size; goto fin;}
-	if (l == m) {m = r; break;}
-	if ((t = (*cmp)(l,m,d)) > 0) {eq_r = 0; break;}
-	if (t == 0) break;
+        if ((l += size) == r)
+          {r += size; if (r != m) mmswap(r,m); r += size; goto fin;}
+        if (l == m) {m = r; break;}
+        if ((t = (*cmp)(l,m,d)) > 0) {eq_r = 0; break;}
+        if (t == 0) break;
       }
       mmswap(l,r);    /* swap left and right */
     }
@@ -517,8 +522,8 @@ ruby_qsort(void* base, const size_t nel, const size_t size, cmpfunc_t *cmp, void
     fin:
     if (eq_l == 0)                         /* need to sort left side */
       if (eq_r == 0)                       /* need to sort right side */
-	if (l-L < R-r) {PUSH(r,R); R = l;} /* sort left side first */
-	else           {PUSH(L,l); L = r;} /* sort right side first */
+        if (l-L < R-r) {PUSH(r,R); R = l;} /* sort left side first */
+        else           {PUSH(L,l); L = r;} /* sort right side first */
       else R = l;                          /* need to sort left side only */
     else if (eq_r == 0) L = r;             /* need to sort right side only */
     else goto nxt;                         /* need not to sort both sides */
@@ -539,52 +544,75 @@ ruby_strdup(const char *str)
     return tmp;
 }
 
+#if defined HAVE_GETCWD
+# if defined NO_GETCWD_MALLOC
+
 char *
 ruby_getcwd(void)
 {
-#if defined HAVE_GETCWD
-# undef RUBY_UNTYPED_DATA_WARNING
-# define RUBY_UNTYPED_DATA_WARNING 0
-# if defined NO_GETCWD_MALLOC
-    VALUE guard = Data_Wrap_Struct((VALUE)0, NULL, RUBY_DEFAULT_FREE, NULL);
+    VALUE guard = rb_imemo_tmpbuf_auto_free_pointer();
     int size = 200;
     char *buf = xmalloc(size);
 
     while (!getcwd(buf, size)) {
-	int e = errno;
-	if (e != ERANGE) {
-	    xfree(buf);
-	    DATA_PTR(guard) = NULL;
-	    rb_syserr_fail(e, "getcwd");
-	}
-	size *= 2;
-	DATA_PTR(guard) = buf;
-	buf = xrealloc(buf, size);
+        int e = errno;
+        if (e != ERANGE) {
+            rb_free_tmp_buffer(&guard);
+            rb_syserr_fail(e, "getcwd");
+        }
+        size *= 2;
+        rb_imemo_tmpbuf_set_ptr(guard, buf);
+        buf = xrealloc(buf, size);
     }
+    rb_imemo_tmpbuf_set_ptr(guard, NULL);
+    return buf;
+}
+
 # else
-    VALUE guard = Data_Wrap_Struct((VALUE)0, NULL, free, NULL);
+
+static const rb_data_type_t getcwd_buffer_guard_type = {
+    .wrap_struct_name = "ruby_getcwd_guard",
+    .function = {
+        .dfree = free // not xfree.
+    },
+    .flags = RUBY_TYPED_FREE_IMMEDIATELY | RUBY_TYPED_WB_PROTECTED
+};
+
+char *
+ruby_getcwd(void)
+{
+    VALUE guard = TypedData_Wrap_Struct((VALUE)0, &getcwd_buffer_guard_type, NULL);
     char *buf, *cwd = getcwd(NULL, 0);
-    DATA_PTR(guard) = cwd;
+    RTYPEDDATA_DATA(guard) = cwd;
     if (!cwd) rb_sys_fail("getcwd");
     buf = ruby_strdup(cwd);	/* allocate by xmalloc */
     free(cwd);
+    RTYPEDDATA_DATA(RB_GC_GUARD(guard)) = NULL;
+    return buf;
+}
+
 # endif
-    DATA_PTR(RB_GC_GUARD(guard)) = NULL;
 #else
+
 # ifndef PATH_MAX
 #  define PATH_MAX 8192
 # endif
+
+char *
+ruby_getcwd(void)
+{
     char *buf = xmalloc(PATH_MAX+1);
 
     if (!getwd(buf)) {
-	int e = errno;
-	xfree(buf);
-	rb_syserr_fail(e, "getwd");
+        int e = errno;
+        xfree(buf);
+        rb_syserr_fail(e, "getwd");
     }
-#endif
     return buf;
 }
 
+#endif
+
 void
 ruby_each_words(const char *str, void (*func)(const char*, int, void*), void *arg)
 {
@@ -593,12 +621,12 @@ ruby_each_words(const char *str, void (*func)(const char*, int, void*), void *ar
 
     if (!str) return;
     for (; *str; str = end) {
-	while (ISSPACE(*str) || *str == ',') str++;
-	if (!*str) break;
-	end = str;
-	while (*end && !ISSPACE(*end) && *end != ',') end++;
-	len = (int)(end - str);	/* assume no string exceeds INT_MAX */
-	(*func)(str, len, arg);
+        while (ISSPACE(*str) || *str == ',') str++;
+        if (!*str) break;
+        end = str;
+        while (*end && !ISSPACE(*end) && *end != ',') end++;
+        len = (int)(end - str);	/* assume no string exceeds INT_MAX */
+        (*func)(str, len, arg);
     }
 }
 
